(* $Id: ugraphics.ml,v 1.10 2014/02/27 16:36:36 fclement Exp $ *)

type coordinate = Unatural.t;;

type abscissa = coordinate
and ordinate = coordinate
;;

type center_abscissa = abscissa
and center_ordinate = ordinate
;;

type length = Unatural.t;;

type width = length
and height = length
;;

type aspect_ratio = width * height;;

type radius = length;;

type principal_radius = radius
and secondary_radius = radius
;;

type angle = int;;

type from_angle = angle
and to_angle = angle
;;

type color = int32;;

type color_channel = char;;

type alpha_channel = color_channel
and red_channel = color_channel
and green_channel = color_channel
and blue_channel = color_channel
;;

type from_color = color
and to_color = color
;;

let aspect_ratio_of_int_pair (w, h) =
  let w = Unatural.natural_of_int w in
  let h = Unatural.natural_of_int h in
  (w, h)
;;

let int_pair_of_aspect_ratio (w, h) =
  let w = Unatural.int_of_natural w in
  let h = Unatural.int_of_natural h in
  (w, h)
;;

let int32_of_char c = Int32.of_int (int_of_char c);;

let char_of_int32 i =
  char_of_int (Int32.to_int (Int32.logand i 0x000000ffl));;

let color_channel_of_natural = Unatural.char_of_natural;;
let natural_of_color_channel = Unatural.natural_of_char;;

let a_of c =
  char_of_int32 (Int32.shift_right_logical (Int32.logand c 0xff000000l) 24)
and r_of c =
  char_of_int32 (Int32.shift_right_logical (Int32.logand c 0x00ff0000l) 16)
and g_of c =
  char_of_int32 (Int32.shift_right_logical (Int32.logand c 0x0000ff00l) 8)
and b_of c =
  char_of_int32 (Int32.logand c 0x000000ffl)
;;

let rgb_of_color c = (r_of c, g_of c, b_of c)
;;

let argb a r g b =
  let ( + ) = Int32.add in
  let ( lsl ) = Int32.shift_left in
  int32_of_char a lsl 24 +
  int32_of_char r lsl 16 +
  int32_of_char g lsl 8 +
  int32_of_char b
;;

let rgb = argb '\x00';;

let rgb_int r g b = rgb (char_of_int r) (char_of_int g) (char_of_int b)
;;

open Unatural;;

let red =
  rgb '\xFF' '\x00' '\x00'
;;
let green =
  rgb '\x00' '\xFF' '\x00'
;;
let blue =
  rgb '\x00' '\x00' '\xFF'
;;
let white =
  rgb '\xFF' '\xFF' '\xFF'
;;
let black =
  rgb '\x00' '\x00' '\x00'
;;
let yellow =
  rgb '\xFF' '\xFF' '\x00'
;;
let cyan =
  rgb '\x00' '\xFF' '\xFF'
;;
let magenta =
  rgb '\xFF' '\x00' '\xFF'
;;

type steps = natural
and max_steps = natural
and from_natural = natural
and to_natural = natural;;

(* Scaling from c to c', by i steps over w steps.

 Meaning:
  scale 0 w c c' = c           (1)
  scale w w c c' = c'          (2)
  scale i w c c' = a * i + b   (3) for some real a and b.

  We have  b = c               (due to (1))
  and      c'= a * w + c       (due to (2))
  hence    a = (c' - c) / w

  Finally, we get:

  scale i w c c' = (c' * i + c * (w - i)) / w
*)
let scale i w c c' =
  if is_zero w then c else
  let mid = (w + one) / two in
  let round_quot w x = (x + mid) / w in
  round_quot w (c' * i + c * (w - i))
;;

let round_float f = Unatural.natural_of_float (0.5 +. f)
;;

let scale_natural f n = round_float (f *. Unatural.float_of_natural n)
;;

(* Scaling a color from c to c', by i steps over w steps. *)
let scale_color i w c c' =
  assert (i <= w);
  argb
   (a_of c)
   (color_channel_of_natural
      (scale i w
        (Unatural.natural_of_char (r_of c))
        (Unatural.natural_of_char (r_of c'))))
   (color_channel_of_natural
      (scale i w
        (Unatural.natural_of_char (g_of c))
        (Unatural.natural_of_char (g_of c'))))
   (color_channel_of_natural
      (scale i w
        (Unatural.natural_of_char (b_of c))
        (Unatural.natural_of_char (b_of c'))))
;;

let fill_rectangle x y w h =
  if not (is_zero w) && not (is_zero h) then
    let x = from x
    and y = from y
    and w = from (pred w)
    and h = from (pred h) in
    Graphics.fill_rect x y w h
;;

let fill_rectangle_ball x y rx ry =
  fill_rectangle
    (x - rx) (y - ry) (two * rx + one) (two * ry + one)
;;

let fill_ellipse_ball x y rx ry =
  if is_zero rx || is_zero ry then fill_rectangle_ball x y rx ry else
  let x = from x
  and y = from y
  and rx = from rx
  and ry = from ry in
  Graphics.fill_ellipse x y rx ry
;;

let fill_inscripted_ellipse x y w h =
  if is_odd w && is_odd h then
    let rx = (w - one) / two in
    let ry = (h - one) / two in
    let xc = x + rx in
    let yc = y + ry in
    fill_ellipse_ball xc yc rx ry else
  (* Using round is simply wrong! *)
  failwith "fill_inscripted_ellipse: not yet implemented."
;;

(* To fill balls associated with the L_infinite norm, i.e. the sup norm:
  ||u||_sup = max |u_i|. *)
let fill_ball_inf = fill_rectangle_ball
;;

(* To fill balls associated with the L_2 norm, i.e. the Euclidean norm:
  ||u||_2 = (sum_i |u_i|^2)^(1/2). *)
let fill_ball_euclidean = fill_ellipse_ball
;;

(* Obsolete definitions for backward compatibility.
  Do not use in fresh code. *)
let fill_rect = fill_rectangle;;
let fill_ellipse = fill_ellipse_ball;;


type x = abscissa
and y = ordinate
and xc = center_abscissa
and yc = center_ordinate
and w = width
and h = height
and r = radius
and rx = principal_radius
and ry = secondary_radius
;;

(* Converting to the new types the Graphics basic stuff.
   This should be reimplemented here.
*)
let set_color c =
  Graphics.set_color (Int32.to_int c)
;;

let moveto x y = Graphics.moveto (Unatural.int_of_natural x) (Unatural.int_of_natural y)
;;

let lineto x y = Graphics.lineto (Unatural.int_of_natural x) (Unatural.int_of_natural y)
;;

let set_line_width w = Graphics.set_line_width (Unatural.int_of_natural w)
;;

let draw_circle x y r =
  Graphics.draw_circle
    (Unatural.int_of_natural x) (Unatural.int_of_natural y) (Unatural.int_of_natural r)
;;

let draw_arc x y rx ry a1 a2 =
  Graphics.draw_arc
    (Unatural.int_of_natural x) (Unatural.int_of_natural y)
    (Unatural.int_of_natural rx) (Unatural.int_of_natural ry)
    a1 a2
;;

let plot x y = Graphics.plot (Unatural.int_of_natural x) (Unatural.int_of_natural y)
;;

let open_graph = Graphics.open_graph;;

let fill_poly a =
  Graphics.fill_poly
    (Array.map (fun (x, y) -> int_of_natural x, int_of_natural y)
     a)
;;

let fill_circle xc yc r =
  Graphics.fill_circle
    (int_of_natural xc)
    (int_of_natural yc)
    (int_of_natural r)
;;

